# -*- coding: utf-8 -*-
# flake8:noqa

# %% [markdown]
## Demo 1): First steps and the Field class
# In this first demo notebook, you'll learn how to import the `empyre` package and how to utilize the `Field` class,
# which can contain multidimensional vector and scalar fields and is core to many functions of `empyre`. If you haven't
# already, please visit the [documentation](https://empyre.iffgit.fz-juelich.de/empyre/) to see how to install `empyre`.


# %% [markdown]
### Import:
# First import `empyre`, we recommend to use the abbreviation `emp`. We'll numpy as well.
# %%
import empyre as emp
import numpy as np


# %% [markdown]
### Represent your data with the 'Field' class
# A `Field` object is a container for multi-dimensional vector or scalar fields, each Field has three parameters:
#
# * **data**: Your underlying data, *e.g.* as a numpy array.
#
# * **scale**: The axis scaling along each axis, given as a tuple. `scale` can be a given as a single number during
# construction of the `Field` object, which assumes that the scale is the same for all axes.
#
# * **vector**: Determines if the field is a scalar field (`vector=False`, which is the default) or a vector field
# (`vector=True`). Please note that, if `vector=True`, the last axis of `data` is interpreted as the dimension of the
# vector components. The component dimension is the last one, so that the `Field` class can be indexed similarily to
# numpy arrays. This way, indexing a field with just spatial indices (by dropping the trailing component dimension)
# will work on scalar and vector fields at the same time. Most functions of `Field` are designed to work on both
# vector and scalar fields, as well.
#
#
# We'll start with a 2D scalar and a 2D vector field. You'll find the base `Field` class in the `fields` submodule.
# Note that you can print fields to get a short summary about their parameters.
# %%
scalar_data = np.zeros((16, 16))
scalar_field = emp.fields.Field(data=scalar_data, scale=1, vector=False)
print(scalar_field)

vector_data = np.zeros((16, 16, 3))
vector_field = emp.fields.Field(data=vector_data, scale=1, vector=True)
print(vector_field)


# %% [markdown]
### Create more interesting fields with `create_...` functions
# The two fields created before are empty and not very interesting. The `fields` submodule has a few alternate
# constructor functions to generate shapes and vector fields in particular.
# Let's look at the `shapes` submodule first, whose documentation can be found
# [here](https://empyre.iffgit.fz-juelich.de/empyre/fields.html#module-empyre.fields.shapes) (Note that all functions
# can be directly accessed via `emp.fields.create_...`, no need to access `shapes` directly).
# A simple example is the disc shape:
# %%
field_disc = emp.fields.create_shape_disc(dim=(32, 32), radius=10)
print(field_disc)


# %% [markdown]
### Visualize fields with the 'vis' subpackage
# The `vis` subpackage is used to visualize field objects. Its interface mirrors that of matplotlib and should be easy
# to learn if you are familiar with the latter. You can find the documentation
# [here](https://empyre.iffgit.fz-juelich.de/empyre/vis.html).
# Usually, scalar fields can be visualised with the `imshow` function (additional *kwargs* will be forwarded to the
# underlying matplotlib function, a behaviour shared by almost all 2D plot functions in `vis`):
#
# %%
emp.vis.imshow(field_disc)



# %% [markdown]
### Some `vis` basics
# The `vis` package has a lot of decorators that you can use to enhance your plots, you can find the documentation
# [here](https://empyre.iffgit.fz-juelich.de/empyre/vis.html#module-empyre.vis.decorators).
# To apply them to a plot, just use the commands in the lines following a plot command. To generate a new plot, use the
# `vis.new()` function. If you are in a Jupyter notebook, like here, you don't need `new` at the beginning of each cell,
# but `new` has some other neat features (like returning a grid of subplots) that could be useful for your specific
# case. Furthermore, `empyre` uses custom matplotlib stylesheets for 'images' and 'plots', respectively, which you can
# explicitely set in the `new` function with the `mode` parameter. For publications, `new` can specify the exact size
# of your plots according to your LaTeX document and save them afterwards with the `savefig` function.
# See [here](https://empyre.iffgit.fz-juelich.de/empyre/vis.html#empyre.vis.tools.new) for more info on all of those
# features. `empyre` provides mainly plotting routines for images, but you can mix and match with matplotlib routines,
# for plots.
# %%
emp.vis.new(mode='image')  # Not necessary here, just for clarity, 'image' is the default!
emp.vis.imshow(field_disc)
emp.vis.scalebar()
# Create a new figure with two plots
fig, axes = emp.vis.new(nrows=1, ncols=2, aspect=0.3, mode='plot')
x = np.arange(10)
axes[0].plot(x, np.random.rand(10))
axes[1].plot(x, np.random.rand(10))

# %% [markdown]
### Creation of vector fields
# The `fields` module can not only create shapes but also some special vector fields like homogeneous distributions
# (all vectors point in the same direction) or vortices / skyrmions, all of which you can find
# [here](https://empyre.iffgit.fz-juelich.de/empyre/fields.html#module-empyre.fields.vectors).
# The according functions all start with `emp.vis.create_vector_...`.
# `vis` also has special plotting functions for vector fields, often with directional color encoding, where the vectors
# are colored according to:
# * **hue**: encodes the in-plane direction of the vector.
#
# * **brightness**: encodes the out-of-plane component of the vector (light: up, dark: down)
#
# * **saturation**: encodes the amplitude of the vector with central gray corresponding to an amplitude of zero.
#
# One example for such a plot is the `colorvec` funtion. You can always add the colorwheel to your plots with the
# `colorwheel` function, as well. Let's try all of this out with a vortex vector distribution with an out-of-plain core
# with a radius of 3 pixels, pointing upwards.
# %%
field_vortex = emp.fields.create_vector_vortex(dim=(32, 32), core_r=3, oop_r=5)
print(field_vortex)
# And plot (colorwheel shows directional color encoding):
emp.vis.colorvec(field_vortex)
emp.vis.scalebar()
emp.vis.colorwheel()


# %% [markdown]
### Quiver plot
# Another way of plotting vector fields is provided by the `quiver` plot function (using the matplotlib function of the
# same name under the hood). `quiver` automatically reduces the number of drawn arrows for clarities sake. Have a look
# into the function documentation to learn more.
# %%
emp.vis.quiver(field_vortex)
emp.vis.scalebar()
emp.vis.colorwheel()


# %% [markdown]
### Combining vector and shapes
# The `Field` class implements many basic math operators like `+`, `-` or `*`. The latter can be used to combine a
# constructed vector field and a shape (that acts like a mask in this case). Furthermore, you can combine `colorvec` and
# `quiver` plots to create appealing images.
# %%
field_comb = field_disc * field_vortex
emp.vis.colorvec(field_comb)
emp.vis.quiver(field_comb)
emp.vis.scalebar()
emp.vis.colorwheel()


# %% [markdown]
### Congratulations
# You managed to create and plot your first 2D vector fields! Look at the functions docstrings or the documentation of
# `empyre` to learn more or have a look at the next demos to find out how to handle 3D fields and more functionality of
# the `Field` class.